/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/*
 * $Id: cred.c,v Exp
 *
 * $Log: cred.c,v $
 * Revision 1.7.4.2  2002/11/13 07:42:59  mcnabb
 * Remove extraneous functions and ifdef's for 2.4.18-17
 *
 * Revision 1.7.4.1  2002/11/11 18:21:43  mcnabb
 * As part of kernel changes in 2.4.18-17, sys_call_table is no longer exported.  This will require a change in NSS and SMB code
 *
 * Revision 1.7  2001/12/14 21:04:07  gjertsen
 * Fix NAS code for IA64 (so it at least compiles).
 * Ensure that plat headers are only included in associated cxi headers.
 *
 */

#ifdef CONFIG_NSS

#include <Shark-gpl.h>

#include <linux/types.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <cxiTypes.h>
#include <cxiCred.h>

static int cred_setgroups32(int gidsetsize, gid_t *grouplist)
{
  gid_t top[2];
	
  if (HAS_PAG(current->groups)) {
    if (gidsetsize < 2)
      goto override_setgroups;
    if (copy_from_user(top, grouplist, 2 * sizeof(gid_t)))
      return -EFAULT;
    if (HAS_PAG(top))
      goto old_setgroups;
  } else {
    goto old_setgroups;
  }

 override_setgroups:
  if (!capable(CAP_SETGID))
    return -EPERM;
  if (((unsigned)gidsetsize) + 2 > NGROUPS)
    return -EINVAL;
  if(copy_from_user(&current->groups[2], grouplist, gidsetsize * sizeof(gid_t)))
    return -EFAULT;
  current->ngroups = gidsetsize + 2;
  return 0;
  
 old_setgroups:
  LOGASSERT(0);
  return -ENOSYS;
}

#if !defined(GPFS_ARCH_POWER) && !defined(GPFS_ARCH_IA64)
static int cred_setgroups16(int gidsetsize, old_gid_t *grouplist)
{
  old_gid_t top[2];
  old_gid_t gids[NGROUPS-2];
  int i;

  if (HAS_PAG(current->groups)) {
    if (gidsetsize < 2)
      goto override_setgroups;
    if (copy_from_user(top, grouplist, 2 * sizeof(old_gid_t)))
      return -EFAULT;
    if (HAS_PAG(top))
      goto old_setgroups;
  } else {
    goto old_setgroups;
  }

 override_setgroups:
  if (!capable(CAP_SETGID))
    return -EPERM;
  if (((unsigned)gidsetsize) + 2 > NGROUPS)
    return -EINVAL;
  if (copy_from_user(gids, grouplist, gidsetsize * sizeof(old_gid_t)))
    return -EFAULT;
  for (i = 0; i < gidsetsize; i++)
    current->groups[i+2] = (gid_t)gids[i];
  current->ngroups = gidsetsize + 2;
  return 0;
  
 old_setgroups:
  LOGASSERT(0);
  return -ENOSYS;
}
#endif

#endif /* CONFIG_NSS */
