/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/*
 * Implementation of shared segment for GPFS daemon and GPFS kernel code.
 *
 * Contents:
 *
 * $Id: ss_ia64.c,v 1.30.4.1 2002/09/06 15:49:10 mcnabb Exp $
 *
 * $Log: ss_ia64.c,v $
 * Revision 1.30.4.1  2002/09/06 15:49:10  mcnabb
 * Fix build break.
 *
 * Revision 1.31  2002/08/13 20:27:55  gjertsen
 * Fix build break.
 *
 * Revision 1.30  2002/07/26 23:29:35  dcraft
 * Make AIX 32/64 bit shared memory creation common under NEWBM define.
 * Make NEWBM code compile under all platforms.  Fix shutdown
 * under NEWBM code.  Should be no functional change outside
 * NEWBM ifdef.
 *
 * Revision 1.29  2001/10/30 16:24:43  gjertsen
 * Add macros for flush tlb operations. Optimize IA64 code to more
 * selectively flush the tlb cache.
 *
 * Revision 1.28  2001/10/08 17:22:24  eshel
 * move kxSaveThreadInfo() for ia64 to ss_ia64.c
 *
 * Revision 1.27  2001/09/22 20:10:02  dcraft
 * Remove kiobufs from cxiKernelIODescriptor_t.  Use temporary
 * kiobufs for map/unmap.   Remove dead code and dead comments
 * in portability layer and update readmes and license.
 * Fix traceback to appear in mmfs.log file.
 *
 * Revision 1.26  2001/09/21 22:54:57  yuri
 * Remove core dump code.
 *
 * Revision 1.25  2001/09/19 23:40:37  eshel
 * Add privilege level support for linux ia64.
 *
 * Revision 1.24  2001/08/09 21:11:42  dcraft
 * Modifications to allow running on latest Redhat 7.1 update
 * Kernel version 2.4.3-12.
 * Requires checkout of new site.mcr.proto
 *
 * Revision 1.23  2001/06/14 18:12:21  gjertsen
 * Initial changes for IA64 beta RH 7.1 with 2.4.3-3 kernel
 *
 * Revision 1.22  2001/05/25 14:49:53  gjertsen
 * Minor fixes to get IA64 code to compile again. Temporarily
 * ignore broken traceback and core dump code.
 *
 * Revision 1.21  2001/04/21 01:04:58  wyllie
 * Use new infrastructure for allocating and making accessible memory shared
 * between the kernel and the daemon.
 *
 * Revision 1.20  2001/03/06 21:55:39  dcraft
 * Minor compatible mod for building against linux-2.4.2-ac12
 *
 * Revision 1.19  2001/02/26 17:33:29  dcraft
 * Pare down unused code pieces.  Run cindent.
 *
 * Revision 1.18  2001/01/28 20:43:49  dcraft
 * Invoke lcrash for kernel thread tracebacks.  Mod kxDumpTraceBack()
 * to allow a file name for output.  Normal ksymoops is backup if
 * lcrash isn't present.
 *
 * Revision 1.17  2000/12/20 22:19:35  gjertsen
 * Update copyright.
 *
 * Revision 1.16  2000/12/12 16:53:43  eshel
 * Add code to support trace back for ia64.
 *
 * Revision 1.15  2000/12/08 20:39:58  eshel
 * don't use CONFIG_IA64_NEW_UNWIND on ia64 for now
 *
 * Revision 1.14  2000/12/06 20:20:02  gjertsen
 * Add in some changes needed with new IA64 kernel.
 *
 * Revision 1.13  2000/12/05 23:41:17  eshel
 * Fix trace back for ia64.
 *
 * Revision 1.12  2000/12/05 22:07:59  gjertsen
 * IA64 fixes for latest gnu toolchain w/ glibc2.2 and 2.4.0test10 kernel.
 *
 * Revision 1.11  2000/12/05 19:13:00  eshel
 * Add code for core dump on ia64.
 *
 * Revision 1.10  2000/11/30 22:11:45  eshel
 * Add and move code for ia64.
 *
 * Revision 1.9  2000/11/30 21:02:00  gjertsen
 * Add platform specific headers for kernel extension functions to help avoid
 * further #ifdef proliferation. Minor IA64 build fix.
 *
 * Revision 1.8  2000/11/17 15:22:17  gjertsen
 * Rename or make static several functions.
 * Toggle off kdb stuff on IA64 for now.
 *
 * Revision 1.7  2000/11/10 21:06:08  eshel
 * Fix comments.
 *
 * Revision 1.6  2000/11/10 17:21:00  eshel
 * Add ia64 code for trace back.
 *
 * Revision 1.5  2000/11/09 17:40:48  gjertsen
 * Misc cleanup to allow module load.
 *
 * Revision 1.4  2000/11/09 01:32:12  eshel
 * Move more code to common routines for ia64 and i386.
 *
 * Revision 1.3  2000/11/08 00:51:24  eshel
 * Fix ia64 compile.
 *
 * Revision 1.2  2000/11/07 23:01:47  eshel
 * Move more common code to ss.c.
 *
 * Revision 1.1  2000/11/07 18:44:01  eshel
 * Start moving common code from ia64/ss.c and i386/ss.c to parent directory.
 *
*/

#include <Shark-gpl.h>
#include <Logger-gpl.h>
#include <cxiSystem.h>
#include <cxiIOBuffer.h>
#include <cxiSharedSeg.h>
#include <Trace.h>

#include <linux/types.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/binfmts.h>
#include <linux/vmalloc.h>
#include <linux/signal.h>

#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/user.h>
#include <asm/mman.h>
#include <asm/atomic.h>
#include <asm/ptrace.h>

#include <arch-gpl.h>

#ifdef LEVEL_1
struct ia64_pfs {
	__u64 reserved0 : 62;
	__u64 ppl : 2;
};
# define ia64_pfs(regs) 	((struct ia64_pfs *) &(regs)->ar_pfs)
#endif

int
set_privilege_level(unsigned long progLevel)
{
#ifdef LEVEL_1
	unsigned long cpl;
	unsigned long new_cpl;
	unsigned long ipsr;
	unsigned long new_ipsr;
	unsigned long pfs;
	unsigned long new_pfs;
        struct pt_regs *regs;

        if (progLevel < 0 || progLevel > 3 || !cxiIsSuperUser())
          return -1;

        regs = ia64_task_regs(current);
        cpl = ia64_psr(regs)->cpl;
        ipsr = regs->cr_ipsr;
        pfs = regs->ar_pfs;

        ia64_psr(regs)->cpl = progLevel; // set current privilege level
        ia64_pfs(regs)->ppl = progLevel; // set previous privilege level

        new_ipsr = regs->cr_ipsr;
        new_pfs = regs->ar_pfs;

        TRACE5(TRACE_SHARED, 2, TRCID_SS_022,
               "set_privilege_level: cpl 0x%lX ipsr 0x%lX pfs 0x%lX new_ipsr 0x%lX new_pfs 0x%lX\n",
               cpl, ipsr, pfs, new_ipsr, new_pfs);

#endif
	return 0;
}

int
get_privilege_level()
{
	unsigned long cpl;
        struct pt_regs *regs;
	unsigned long ipsr;
	unsigned long pfs;

        regs = ia64_task_regs(current);
        cpl = ia64_psr(regs)->cpl;
        ipsr = regs->cr_ipsr;
        pfs = regs->ar_pfs;

        TRACE3(TRACE_SHARED, 2, TRCID_SS_022X,
               "get_privilege_level: cpl 0x%lX ipsr 0x%lX pfs 0x%lX\n",
               cpl, ipsr, pfs);

	return cpl;
}

int
kxSaveThreadInfo(int tid, void* regP)
{
  struct sigcontext *sc = (struct sigcontext *)regP;
  struct task_struct *p;
  struct pt_regs *pt;
  int rc = ENOENT;

  for_each_task(p)
  {
    if (current->pgrp == p->pgrp && current->pid != tid && p->pid == tid)
    {
      pt = ia64_task_regs(p);

      rc = __put_user(pt->cr_iip, &sc->sc_ip);
      rc |= __put_user(pt->ar_bspstore, &sc->sc_ar_bsp);
      rc |= __put_user(pt->ar_pfs, &sc->sc_cfm);

      return rc;
    }
  }
  return -1;
} 	

/* TLB flush routines for a single page or an address range.
   These are not exported by the kernel (like flush_tlb_all)
   and we want to do this at a finer granularity than flushing
   the entire TLB cache.
   The kernel serializes tlb cache flushes with the ptcg_lock,
   but we don't have access to this. */

void flush_tlb_page_ia64(unsigned long start_addr)
{
  unsigned long addr = start_addr & ~(PAGE_SIZE - 1);  /* page align */

  /* ptc commands expect flush size starting at bit 2 of register */
#ifdef CONFIG_SMP
  __asm__ __volatile__ ("ptc.ga %0, %1;;"
                        :
                        : "r"(addr),
                          "r"(PAGE_SHIFT << 2)
                        : "memory");
#else
  __asm__ __volatile__ ("ptc.l %0, %1;;"
                        :
                        : "r"(addr),
                          "r"(PAGE_SHIFT << 2)
                        : "memory");
#endif /* CONFIG_SMP */  

  /* need to serialize */
  __asm__ __volatile__ ("srlz.i;;" : : : "memory");
}

void flush_tlb_range_ia64(unsigned long start_addr, unsigned long end_addr)
{
  unsigned long addr = start_addr & ~(PAGE_SIZE - 1);  /* page align */

  while (addr < end_addr)
  {
    /* ptc commands expect flush size starting at bit 2 of register */
#ifdef CONFIG_SMP
    __asm__ __volatile__ ("ptc.ga %0, %1;;"
                          :
                          : "r"(addr),
                            "r"(PAGE_SHIFT << 2)
                          : "memory");
#else
    __asm__ __volatile__ ("ptc.l %0, %1;;"
                          :
                          : "r"(addr),
                            "r"(PAGE_SHIFT << 2)
                          : "memory");
#endif /* CONFIG_SMP */  

  /* need to serialize */
  __asm__ __volatile__ ("srlz.i;;" : : : "memory");

    addr += PAGE_SIZE;
  }
}
