/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
#ifndef _h_cxiAclUser
#define _h_cxiAclUser
/* @(#)02       1.3  src/avs/fs/mmfs/ts/kernext/ibm-kxi/cxiAclUser.h, mmfs, avs_rmar, rmart1f7 7/30/02 10:28:19 */
/****************************************************************************
 *
 * Acl User data
 *
 * Contains Tiger Shark Access Control List definitions.
 * Some of the definitions in this file are exportable data items,
 * if POSIX compliance is claimed.
 *
 ***************************************************************************/
/*
 * $Id: cxiAclUser.h,v 1.3 2002/07/30 15:00:35 esj Exp $
 *
 * $Log: cxiAclUser.h,v $
 * Revision 1.3  2002/07/30 15:00:35  esj
 * Defect 368861:  Rename ACL_TYPE_ACCESS to GPFS_ACL_TYPE_ACCESS and
 * ACL_TYPE_DEFAULT to GPFS_ACL_TYPE_DEFAULT to be consistent with other
 * externalized constants.  Restore the original values in cxiAclUser.h
 *
 * Revision 1.2  2002/07/30 01:57:54  esj
 * Defect 368861:  Move the definitions of ACL_TYPE_ACCESS and ACL_TYPE_DEFAULT
 *                 to gpfs.h
 *
 * Revision 1.1  2002/01/10 19:17:54  eshel
 * Move some defines from AclUser.h to a cxi file.
 *
 *
 */

#include <cxiCred.h>

/* An ACL entry's permission set is defined as a set of 8 bits.
   Bits seven and eight are currently unused.
   Bits six, five, and four are delete, insert, and control access.
   Bits three, two, and one are read, write, and execute/search access.
   If any of these bits are set, the corresponding access is granted.
   Otherwise, access is denied.
   Note: This data type is non-exportable data. */
typedef unsigned char acl_permset_t;

/* This data type is used to distinguish different types of ACLs.
   Note: This data type is exportable data. */
typedef unsigned char  acl_type_t;
#define ACL_TYPE_ACCESS            1
#define ACL_TYPE_DEFAULT           2

// The following are for internal use only
#define ACL_TYPE_ACCESS_DELETE   252
#define ACL_TYPE_DEFAULT_DELETE  253
#define EXTENDED_ACCESS_ONLY     254
#define INTERNAL_FORMAT          255

/* acl_permset_t bits
 */
#define ACL_PERM_EXECUTE  001
#define ACL_PERM_WRITE    002
#define ACL_PERM_READ     004
#define ACL_PERM_CONTROL  010   /* Version 2 ACL permission only     */
#define ACL_PERM_INSERT   020   /* Version 2 ACL permission only     */
#define ACL_PERM_DELETE   040   /* Version 2 ACL permission only     */

#ifdef CONFIG_NSS
/*
 * Meaning of "insert" depends on type - for dirs, it means adding files,
 * for files, it means overwriting file data.
 * Meaning of "append" depends on type - for dirs, it means adding subdirs,
 * for files, it means append-only writes.
 * 
 * Note: acl_permset_t is a UChar, so not all ACL_PERM_xx bits below can be
 * used. Currently, ACL_PERM_STAT gets mapped to ACL_PERM_READ (NT) or 
 * ACL_PERM_EXECUTE (Unix) depending on platform access.
 */
#define ACL_PERM_APPEND  0100   /* Permission to append files, create subdir */
#define ACL_PERM_GETACL  0200   /* Permission to read file ACL */
#define ACL_PERM_STAT    0400   /* Permission to stat files */
#define ACL_PERM_MASK    0777   /* Mask includes all permissions     */
/* 
 * An NT caller requires ACL_PERM_READ on dir for lookups on files and subdirs
 * Unix caller requires ACL_PERM_EXECUTE
 */
#define ACL_PERM_LOOKUP(credP) (cxiIsNTUser(credP) ? ACL_PERM_STAT : ACL_PERM_EXECUTE)
#else
#define ACL_PERM_LOOKUP(credP) ACL_PERM_EXECUTE
#define ACL_PERM_MASK     077   /* Mask includes all permissions     */
#endif

#define ACL_PERM_UMODE    007   /* Mask includes only UNIX mode bits */

/* VNOP extended access modes that can be checked (gpfs_v_access()).
 */
#define C_ACC   ACL_PERM_CONTROL
#define I_ACC   ACL_PERM_INSERT
#define D_ACC   ACL_PERM_DELETE

/* An ACL entry's tag type is defined as a one byte value.
   It is used to distinguish different types of ACL entries.
   Note: This data type is exportable data. */
typedef unsigned char  acl_tag_t;
#define ACL_USER_OBJ        1   /* aceShort ACL entry */
#define ACL_GROUP_OBJ       2   /* aceShort ACL entry */
#define ACL_OTHER           3   /* aceShort ACL entry */
#define ACL_MASK            4   /* aceShort ACL entry */
#define ACL_USER            5   /* aceShort ACL entry */
#define ACL_GROUP           6   /* aceShort ACL entry */
#define ACL_FOREIGN_OTHER   7   /* Version 2 ACL entry type (use aceLong)  */
#define ACL_ANY_OTHER       8   /* Version 2 ACL entry type (use aceShort) */
#define ACL_FOREIGN_USER    9   /* Version 2 ACL entry type (use aceLong)  */
#define ACL_FOREIGN_GROUP  10   /* Version 2 ACL entry type (use aceLong)  */
#define ACL_UDEFINED_TAG  255

struct acenodce
 {
  char        aceComments[1];  /* Short ACL entry                  */
   };
struct acedce
 {
  ext_id_t    aceRealm;        /* DCE cell specification           */
  char        aceComments[1];  /* Version 2 ACL foreign ACL entry  */
 };


/* An ACL entry consists of a fixed part - ACL entry length, permission set,
   entry tag type, entry qualifier (uid or gid), and followed by an optional
   character string.
   Note: This data type is non-exportable data. */
typedef struct
{
  UInt16          aceLength;
  acl_permset_t   acePermissions;
  acl_tag_t       aceTagType;
  cxiUid_t        aceQualifier;
  union
  {
    struct acenodce  aceShort;
    struct acedce aceLong;
  };
} acl_entry_t;

#define SHORT_ACE_SIZE PTR_TO_INT32(&(((acl_entry_t *) 0)->aceShort.aceComments[0]))
#define LONG_ACE_SIZE  PTR_TO_INT32(&(((acl_entry_t *) 0)->aceLong.aceComments[0]))

struct aclv1
{
  acl_entry_t   aclEntry[1];
};

struct aclv2
{
  unsigned int  reserved   : 7;    // unused guaranteed to be zero
  unsigned int  aclExtMode : 9;    // Internal use of CID permission
  UInt16        aclOtherAceCnt;    // number of any_other & foreign_other
  ext_id_t      defaultRealm;      // Default ACL cell
  acl_entry_t   aclEntry[1];       // zero or more ACL entries
};

/* An ACL consists of header information (length, type, etc.).
   followed by at least three, but may be more, ACL entries.
   The three required entries must be ACL_USER_OBJ, ACL_GROUP_OBJ,
   and ACL_OTHER. The first four bytes of the ACL header must always
   be the length of the entire ACL.
   Note: This data type is non-exportable data. */
typedef struct
{
  int               aclLength;          // total length (hdr + acl entries)
  UInt16            aclVersion;         // version number
  acl_type_t        aclType;            // type of ACL: default or access
  union
  {
    char            aclBuf[1];          // Opaque NT ACL used by NSS
    acl_permset_t   aclMask;            // ACL_MASK entry
    acl_permset_t   aclGroupObj;        // ACL_GROUP_OBJ entry
  };
  cxiMode_t         aclMode;            // file mode (at acl create time)
  int               aclUserAceCnt;      // number of ACL_USER entries
  int               aclGroupAceCnt;     // number of ACL_GROUP entries
  unsigned int      aclNewModeAttr : 1; // if set, replace the current mode
  unsigned int      reserved : 31;      // guaranteed to be zero
  union
  {
    struct aclv1  v1;
    struct aclv2  v2;
  };
} ACL;


/* The following is used as a pointer to an ACL in working storage.
   Note: This data type is non-exportable data. */
typedef ACL *acl_t;

#define ACL_HEADER_SIZE_V1   PTR_TO_INT32(&(((acl_t) 0)->v1.aclEntry[0]))
#define ACL_HEADER_SIZE_V2   PTR_TO_INT32(&(((acl_t) 0)->v2.aclEntry[0]))

#define ACL_ENTRY_START(ACLP,VER) \
          (acl_entry_t *)((VER == ACL_FIRST_VERSION) ? \
            (caddr_t)ACLP + ACL_HEADER_SIZE_V1 :  \
            (caddr_t)ACLP + ACL_HEADER_SIZE_V2);

/* Upper bound on the size of a base acl (header + 4 entries for base
   permissions and mask).  The correct value would be
     ACL_HEADER_SIZE + 4*SHORT_ACE_SIZE
   However, ACL_BASE_SIZE_MAX needs to be a compile-time constant, because
   it is used to declare a char array on the stack as temporary buffer for
   assembling a base acl.  Unfortunately, the compiler does not accept the
   offset calculations in the ACL_HEADER_SIZE and SHORT_ACE_SIZE macros as
   constants.  The definition below slightly over-estimates the base acl size,
   but is sufficient for our purpose, since we only need an upper bound. */
#define ACL_BASE_SIZE_MAX (sizeof(ACL) + 3*sizeof(acl_entry_t))

/* acl format versions */
#define ACL_FIRST_VERSION     1
#define ACL_SECOND_VERSION    2
#define ACL_VERSION           2
#ifdef CONFIG_NSS
#define ACL_NT_VERSION        3
#define ACL_HEADER_SIZE_NT    PTR_TO_INT32(&(((acl_t) 0)->aclBuf[0])) 
#endif

#define DEFAULT_ACLMASK_V2    077       // rwxcid
#define DEFAULT_ACLMASK_V1    07        // rwx

#define ACLMASK(_VERSION) \
          (((_VERSION) == ACL_FIRST_VERSION) ? DEFAULT_ACLMASK_V1 :  \
                                               DEFAULT_ACLMASK_V2);
#define MASK_SET           0x80

/* Mapping for the index tables used by the verifyTSAcl routine.
   Note: This data type is non-exportable data. */
typedef struct
{
  acl_entry_t     *aceP;
  cxiUid_t         key;
  ext_id_t         extKey; /* cell */
  caddr_t          nxt;
} aceIndxTable_t;

/* Macro to test whether the ACL_MASK is explicitly set */
#define ACLMASK_SET(_aclMask) (((_aclMask) & MASK_SET) != 0)

#endif /* _h_cxiAclUser */

