/***************************************************************************
 *
 * Copyright (C) 2002 International Business Machines
 * All rights reserved.
 *
 * This file is the source for the liblgm.so interface library to 
 * Myricom's GM library.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* $Id: lgm.c,v 1.1 2002/01/20 15:14:10 blackmor Exp $
 *
 *
 * $Log: lgm.c,v $
 * Revision 1.1  2002/01/20 15:14:10  blackmor
 * Open source interface to the Myricom GM library. To be checked in for
 * post emperor release.
 *
 */


#include <pthread.h>
#include <malloc.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include "gm.h"
#include "lgm.h"
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <fcntl.h>
#include <sys/uio.h>

unsigned int port_id=2;
static struct gm_port *p;
static gm_u32_t gm_port_num=2, gm_device=0;
static char     *gm_port_name="LAPI/GM";     /* GM NAME */
static gm_alarm_t  lapi_hal_alarm;

#define LOGPAGE 12

/* All messages will be treated as if they are a datagram sized message for now */
int
find_buf_size(int count)
{
     return LOGPAGE;
}

/* Wrapper function for gm_datagram_send */ 
void
gm_lapi_send(void *bufi, int count, int taskid, void *callback)
{
    int size;
    
    size = find_buf_size(count);
     gm_datagram_send(p,(void *)bufi,size,count,
                 GM_HIGH_PRIORITY, taskid, port_id,
                 callback, (void *) bufi);
}

/* Wrapper function for providing a receive buffer */
void
lapi_provide_buf(void *bufc)
{
       gm_provide_receive_buffer_with_tag(p,(void *) bufc,
       (int)  *( (char *) bufc),GM_HIGH_PRIORITY, *( (char *) bufc+1));
}


/* dedicated thread will call process GM events in a loop 
 * Call with a lock held that is only released when sleeping 
 * Callback function is provided for processing received buffers.
 */
void
lapi_gm_event_loop(pthread_mutex_t *dev_lck, lapi_gm_updatebuf_t *cb, void *cb_param)
{
        union gm_recv_event *e;
        char *bufc;

        e= gm_blocking_receive_no_spin(p);
        while( GM_N(e->recv.type) != GM_NO_RECV_EVENT ) { 
              switch ( GM_N(e->recv.type)) {

                   case GM_RECV_EVENT:
                   case GM_HIGH_RECV_EVENT:
                   bufc = (char *) ntohl(GM_N(e->recv.buffer));

                            
                   if(((void *)ntohl(GM_N(e->recv.message)) != NULL ) &&
                      (ntohl(GM_N(e->recv.message)) !=
                       ntohl(GM_N(e->recv.buffer)))) {

                   gm_memorize_message((void *) ntohl(GM_N(e->recv.message)),
                        (void *) ntohl(GM_N(e->recv.buffer)),
                        (unsigned int)  ntohl( GM_N(e->recv.length)));
                   }

                   /* save tag, length,  size and source */
                   *bufc = (char)GM_N(e->recv.size);
                   *(bufc+1) = (char) GM_N(e->recv.tag);
                   ((short *) bufc)[2] = 0;
                   gm_assert( ((short *) bufc)[3] == 
                           (short) ntohl( GM_N(e->recv.length))-8);

                   (*cb)(cb_param, (void *) bufc);
                        break;

                   case _GM_SLEEP_EVENT:
                        pthread_mutex_unlock(dev_lck);
                        gm_unknown(p,e);
                        pthread_mutex_lock(dev_lck);
                        break; 


                   default:
                        gm_unknown(p,e);
                        break;
              }
              e= gm_receive(p);
          }
          return;
}

/* wrapper function for resetting the alarm */
void lapi_reset_alarm(int time, void *callback, void *param)
{
      gm_cancel_alarm(&lapi_hal_alarm);
      gm_set_alarm(p,&lapi_hal_alarm,time,
         callback, (void *) param);
}

/* wrapper function for setting the alarm */
void lapi_set_alarm(int time, void *callback, void *param)
{
      gm_set_alarm(p,&lapi_hal_alarm,time,
         callback, (void *) param);
}

/* wrapper function for closing GM */
void
lapi_gm_finish_gm()
{
    gm_close(p);
    gm_finalize();
}

/* wrapper function for opening GM */
int 
lapi_gm_open()
{
   int rc;

   if( rc = gm_init()) {
       return rc;
   }

   if( rc = gm_open(&p,gm_device,
                    gm_port_num,gm_port_name, GM_API_VERSION_1_3)) {
       gm_finalize();
       return rc;
   }

  return rc;
}

/* wrapper function for resuming GM after error condition */
void
lapi_gm_resume(void *p, int node_id, void *callback, void *context)
{
        gm_resume_sending(p,GM_HIGH_PRIORITY,node_id,
                          port_id,
                          callback, context);
}
/* wrapper function for obtaining local GM task number */
int
lapi_gm_get_node_id(int *id)
{
   return gm_get_node_id(p,id);
}

/* wrapper function for initializing the GM alarm */
void 
lapi_gm_start_alarm()
{
   gm_initialize_alarm(&lapi_hal_alarm);
}

/* wrapper function for freeing a GM DMAable buffer */
void
lapi_gm_free_buf(void *buf)
{
       gm_dma_free(p,buf);
}

/* wrapper function for finding number of send tokens */
int
get_send_tokens()
{
   return  gm_num_send_tokens(p) - 1;
}
/* wrapper function for finding number of recv tokens */
int
get_recv_tokens()
{
   return  gm_num_receive_tokens(p);
}
/* wrapper function for obtaining DMAable GM  memory */
void *
get_gm_dma_mem(int count)
{
   return  gm_dma_malloc(p,count);
}

/* wrapper function for obtaining limiting message sizes */
void
lapi_set_acceptable_sizes(int high_sizes, int low_sizes)
{
    gm_set_acceptable_sizes( p, GM_HIGH_PRIORITY, high_sizes);
    gm_set_acceptable_sizes( p, GM_LOW_PRIORITY, low_sizes);
}


